from .base import BaseOverlay

config_template = '''router ospf

{}

redistribute kernel
redistribute static
redistribute connected
'''


class OspfOverlay(BaseOverlay):

    protocol = 'OSPF'

    def configure_node(self, node):
        config = config_template.format('\n'.join([
            f'network {s.get_ip_range()} area 0.0.0.0' for s in self._node_subnets(node)
        ]))

        node.activate_protocol(self.protocol, config)



from .. import defaults
from . import subnet as base_subnet


class BaseNode:
    lock_filepath = '/var/run/l3ns.lock'

    def __init__(self, name, internet_connection=False):
        self.name = name
        self._interfaces = {}
        self._routes = {}
        self._networks = []
        self._files = {}

        self.loaded = False
        self.started = False
        self.is_router = False
        self.is_gateway = False
        self.connect_to_internet = internet_connection

    def connect_to(self,
                   other_node: 'BaseNode',
                   subnet_name: str = None,
                   subnet_class: type = None,
                   **subnet_kwargs):

        if subnet_class is None:
            subnet_class = defaults.subnet_class

        if subnet_name is None:
            subnet_name = self.name + '_' + other_node.name

        return subnet_class(subnet_name, self, other_node, **subnet_kwargs)

    def add_interface(self, ip_address, subnet: 'base_subnet.BaseSubnet'):
        self._interfaces[ip_address] = subnet

    def _start(self, *args, **kwargs):
        raise NotImplementedError()

    def start(self, *args, **kwargs):
        self._setup_routes()

        ret = self._start(*args, **kwargs)
        print('node', self.name, 'started')

        self._connect_subnets()
        self._deploy_routes()

        return ret

    def _connect_subnets(self):
        for ip, network in self._interfaces.items():
            self._connect_subnet(network, ip)

    def _connect_subnet(self, subnet, ip):
        raise NotImplementedError()

    def _setup_routes(self):
        max_routes_number = len(self._networks)

        # TODO: LAN routes for routers
        if self.is_router:
            return

        for subnet in self.subnets:
            network = subnet.get_network()

            ip_range = network.ip_range \
                if self.connect_to_internet or network.is_local else 'default'

            if ip_range in self._routes:
                continue

            gateway = subnet.get_gateway(self)
            if gateway:
                self._routes[ip_range] = gateway

            if len(self._routes) == max_routes_number:
                return

    def _deploy_routes(self):
        raise NotImplementedError()

    @property
    def subnets(self):
        return list(self._interfaces.values())

    @property
    def neighbors(self):
        return sum([subnet.nodes for subnet in self.subnets])

    def add_network(self, net):
        self._networks.append(net)

    def load(self):
        raise NotImplementedError()

    # TODO: proper realisation with Subnet optional argument
    def get_ip(self, net=None):
        if net is None:
            return next(iter(self._interfaces.keys()), None)

        else:
            for ip, network in self._interfaces.items():
                if network == net:
                    return ip

            return None

    @classmethod
    def make_router(cls, *args, **kwargs):
        raise NotImplementedError()

    @classmethod
    def generate_nodes(cls, name_prefix, amount, *args, **kwargs):
        return [cls(name_prefix + str(i+1), *args, **kwargs) for i in range(amount)]

    @classmethod
    def generate_routers(cls, name_prefix, amount, *args, **kwargs):
        return [cls.make_router(name_prefix + str(i+1), *args, **kwargs) for i in range(amount)]

    def activate_protocol(self, protocol, config):
        raise NotImplementedError()

    def __repr__(self):
        return '<{class_name}({name}, {ip})'.format(
            class_name=self.__class__.__name__,
            name=self.name,
            ip=self.get_ip())

    def unlock(self):
        # TODO: not the best decision, but we need to avoid starting container before routes are set up
        print(f'unlocking {self.name}')
        self.put_string(self.lock_filepath, '')

    def put_string(self, path, string):
        raise NotImplementedError()



import os
import subprocess
import shutil

from .. import base
from .. import ldc
from .utils import my_hash, generate_wg_keys, find_free_port
from .. import utils

server_config_template = '''[Interface]
ListenPort = {listen_port}
PrivateKey = {private_key}

'''

peer_config_template = '''[Peer]
# {name}
PublicKey = {public_key}
AllowedIPs = {ip_address}/32
'''

server_startup_script = '''
ip link add dev {ifc} type wireguard && \
wg setconf {ifc} {config_file} && \
ip addr add {ip_address}/{net_mask} dev {ifc} && \
ip link set {ifc} up  && \
iptables -A INPUT -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT && \
iptables -A FORWARD -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT && \
iptables -A FORWARD -i {ifc} -o {ifc} -m conntrack --ctstate NEW -j ACCEPT
'''


class WgSubnet(base.BaseSubnet):

    def __init__(self, *args, size=1023, **kwargs):
        super().__init__(*args, size=size + 1, **kwargs)

        self.private_key, self.public_key = generate_wg_keys()
        self.listen_port = None

        # TODO: proper config for this
        self.interface_name = 'wg-' + my_hash(self.name)
        self.config_folder = '/tmp/l3ns/' + self.name
        self.config_filename = os.path.join(self.config_folder, 'wg.conf')

        self._vpn_server_ip = self._get_host_ip()

    def make_server_config(self):
        self.listen_port = find_free_port()
        config = server_config_template.format(
            listen_port=self.listen_port,
            private_key=self.private_key)

        config += '\n\n'.join([peer_config_template.format(
            name=node.name,
            public_key=node.public_key,
            ip_address=node.get_ip(net=self)
        ) for node in self._nodes_dict.values()])

        return config

    def start(self):
        if self.started:
            return

        try:
            os.mkdir(self.config_folder)
        except FileExistsError:
            pass  # TODO: for now

        with open(self.config_filename, 'w') as config_file:
            config_file.write(self.make_server_config())

        ret = subprocess.run(server_startup_script.format(
            ifc=self.interface_name,
            config_file=self.config_filename,
            ip_address=self._vpn_server_ip,
            net_mask=self.prefixlen()
        ),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            shell=True)

        if ret.returncode != 0:
            print('Error while starting subnet ' + self.name + ':', ret.stdout.strip(), sep='\n')

    def load(self):
        # TODO: load keys??
        pass

    def stop(self):
        ret = subprocess.run('ip link delete {}'.format(self.interface_name),
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             universal_newlines=True,
                             shell=True)

        if ret.returncode != 0:
            print('Error while removing subnet ' + self.name + ':', ret.stdout.strip(), sep='\n')

        try:
            shutil.rmtree(self.config_folder)
        except FileNotFoundError:
            print('Error while removing subnet ' + self.name + ':',
                  'WireGuard config directory does not exist', sep='\n')

        print('subnet', self.name, 'stopped')

    @classmethod
    def make_subnet(cls, *args, size=1023, network=None):
        node_list = utils.args.list_from_args(args)

        try:
            cluster_host_set = set([node.cluster_host for node in node_list])
        except AttributeError:
            raise Exception("Only RemoteNodes can be in the ClusterSubnet")

        if len(cluster_host_set) == 1:
            return ldc.DockerSubnet(*args, size=size, network=network)
        else:
            return cls(*args, size=size, network=network)



from .. import ldc
from .node import SwarmNode


class SwarmSubnet(ldc.DockerSubnet):

    def __init__(self, *args, size: int = 510, **kwargs):
        self._cluster_hosts = {}
        self._offset = size
        super().__init__(*args, size=size*2, **kwargs)

    def add_node(self, node: 'SwarmNode'):

        ip_address = self._get_host_ip()

        # docker swarm overlay requires additional ip address for lode balancer each cluster node (cluster_host in l3ns)
        if not isinstance(node, ldc.DockerNode) and node.cluster_host.address not in self._hosts:
            self._cluster_hosts[node.cluster_host.address] = self._hosts.pop(0)

        self._nodes_dict[ip_address] = node
        if node not in self._network:
            self._network.add_node(node)
        node.add_interface(ip_address, self)

    def _get_host_ip(self):
        # Since we don't have control over ip management for lode balancers, first free ip of the docker network
        # ip range will be used for each swarm node. Since it's not possible to control starting order of the nodes,
        # it's easier to just make network twice as large and give out addresses only from the second half
        return str(self._hosts.pop(self._offset+1))

    def start(self):
        if self.started:
            return self.docker_network

        self.docker_network = self._client.networks.create(
            self.name,
            driver='overlay',
            attachable=True,
            ipam=self._make_ipam_config())

        if not self.docker_network.attrs['Driver']:
            raise Exception('Error: IPAM customisation failed for swarm network {}, check if other networks overlap ip pool'.format(self.name))

        self.started = True
        self.loaded = True

        print('[swarm] subnet', self.name, 'started')

        return self.docker_network



from .. import cluster
import subprocess
import re

check_swarm_cmd = 'docker swarm join-token worker'
start_swarm_cmd = 'docker swarm init --advertise-addr {}'
list_nodes_cmd = "docker node ls --format '{{.Hostname}}:{{.Status}}'"


class SwarmHost(cluster.ClusterHost):

    swarm_token = None
    new_swarm = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.swarm_token is None:
            self._init_swarm()

        dc = self.get_docker_client()

        if not self.new_swarm and self._check_if_in_swarm():
            return

        try:
            dc.api.join_swarm([self._get_manager_ip()], self.swarm_token)
        except AttributeError:
            dc.join_swarm([self._get_manager_ip()], self.swarm_token)

    def _check_if_in_swarm(self):
        # check if node in swarm if one exists
        # this cmd gives us hostname of nodes in cluster

        # this can also be used, but hostname'll do for now
        # docker info --format "{{.Swarm.ControlAvailable}}"

        hostname = self.exec_command('hostname')[1].strip()

        ret = subprocess.run(list_nodes_cmd, shell=True, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT, universal_newlines=True)
        if not ret.returncode:
            nodes = [n.split(':') for n in ret.stdout.split('\n')]
            for node in nodes:
                if hostname == node[0] and node[1] == "Ready":
                    return True

        return False


    def _init_swarm(self):
        ret = subprocess.run(check_swarm_cmd, shell=True, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT, universal_newlines=True)

        if ret.returncode:
            print('checking for swarm\nreturn code:', ret.returncode)
            print(ret.stdout)
        else:
            token = self._find_token(ret.stdout)

            if token:
                self.swarm_token = token
                self.new_swarm = False
                return

        ip = self._get_manager_ip()

        ret = subprocess.run(start_swarm_cmd.format(ip), shell=True, stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT, universal_newlines=True)

        if not ret.returncode:
            token = self._find_token(ret.stdout)

            if token:
                self.swarm_token = token
                self.new_swarm = True
                return

        raise Exception("Couldn't start the swarm:\n" + ret.stdout)

    @staticmethod
    def _find_token(s):
        match = re.search('--token ([^ ]*) ', s)

        if match:
            return match.group(1)

        else:
            raise Exception('Error while parsing output for docker swarm token:\n' + s)

    @staticmethod
    def _get_manager_ip():
        # TODO: socket?? hostname?? I dunno
        # the address must be from one of the interfaces, visible from other hosts
        # return '192.168.122.119' # baas test
        return '172.27.27.80'

# new swarm no swarm       - v
# old swarm no swarm       - v
# old swarm has same swarm - v

# new swarm has swarm      - ?
# old swarm has another swarm - ?


if __name__ == "__main__":

    sh = SwarmHost('t2')

    print([n.name for n in sh.get_docker_client().networks.list()])



import docker
import tarfile
import io
import os
import time

from .. import base
from . import utils
from .subnet import DockerSubnet


class DockerNode(base.BaseNode):

    def __init__(self, name, **docker_kwargs):
        self._client = utils.docker_client
        self._docker_kwargs = docker_kwargs
        self.container = None
        self.image = None  # not a name, but docker-py object
        super().__init__(name=name)

    def _connect_subnet(self, subnet, ip):
        subnet.docker_network.connect(self.container, ipv4_address=ip)

    @staticmethod
    def _shell_entrypoint(entrypoint):

        if not entrypoint or type(entrypoint) is str:
            return entrypoint

        else:
            return ' '.join('"{}"'.format(s) for s in entrypoint)

    def _make_entrypoint(self):
        entrypoint = ('entrypoint' in self._docker_kwargs and self._docker_kwargs['entrypoint']) \
                     or self.image.attrs['Config']['Entrypoint']
        cmd = ('command' in self._docker_kwargs and self._docker_kwargs['command']) \
              or self.image.attrs['Config']['Cmd']

        waiting_cmd = 'while [ ! -f {lock_file} ]; do sleep 1; done; '.format(lock_file=self.lock_filepath)

        ret_entrypoint = ['/bin/sh', '-c']

        ret_cmd = waiting_cmd

        if not entrypoint:
            ret_cmd += self._shell_entrypoint(cmd)

        elif type(entrypoint) is str:
            ret_cmd += entrypoint

        elif len(entrypoint) >= 2 and 'sh' in entrypoint[0] and entrypoint[1] == '-c':

            ret_entrypoint = entrypoint[:2]

            ret_cmd += self._shell_entrypoint(entrypoint[2:])

            if cmd:
                ret_cmd += ' ' + self._shell_entrypoint(cmd)

        else:
            ret_cmd += self._shell_entrypoint(entrypoint)

            if cmd:
                ret_cmd += ' ' + self._shell_entrypoint(cmd)

        # for some reason docker messes up while loop if cmd is a string
        return ret_entrypoint, [ret_cmd]

    def _start(self, dc=None):

        if dc:
            self._client = dc
        else:
            dc = self._client

        if self.started:
            return self.container

        networking_kwargs = {}

        try:
            self.image = dc.images.get(self._docker_kwargs['image'])
        except docker.errors.ImageNotFound:
            print('No {} image found locally, trying to pull...'.format(self._docker_kwargs['image']))

            image = self._docker_kwargs['image']

            if ':' in image:
                image, tag = self._docker_kwargs['image'].split(':', maxsplit=2)
            else:
                tag = 'latest'

            self.image = dc.images.pull(image, tag=tag)

        self._docker_kwargs['entrypoint'], self._docker_kwargs['command'] = self._make_entrypoint()

        if 'cap_add' in self._docker_kwargs:
            try:
                self._docker_kwargs['cap_add'].append('NET_ADMIN')
            except AttributeError:
                self._docker_kwargs['cap_add'] = ['NET_ADMIN', self._docker_kwargs['cap_add']]
        else:
            self._docker_kwargs['cap_add'] = 'NET_ADMIN'

        # print(self.name, self._docker_kwargs['entrypoint'] + self._docker_kwargs['command'], sep=': ')

        self.container = dc.containers.run(
            name=self.name,
            detach=True,
            **networking_kwargs,
            **self._docker_kwargs)
        self.started = True
        self.loaded = True

        try:
            default_net = next(iter(self.container.attrs['NetworkSettings']['Networks'].keys()))
        except StopIteration:
            raise Exception('Container {} has no initial net, check docker config')

        if self._interfaces or self.connect_to_internet:
            dc.networks.get(default_net).disconnect(self.container)

        for path, string in self._files.items():
            self.put_string(path, string)

        return self.container

    def load(self, dc=None):

        dc = dc or self._client

        try:
            self.container = dc.containers.get(self.name)
            self.loaded = True
        except docker.errors.NotFound:
            pass

    def stop(self, dc=None):

        dc = dc or self._client

        if not self.loaded:
            self.load(dc=dc)

        try:
            self.container.remove(force=True)
            print('node', self.name, 'stopped')
        except Exception as e:
            print('error while removing node {}: {}'.format(self.name, e))

    def put_string(self, path, string):
        if self.loaded:

            tar = tarfile.TarFile(mode='w', fileobj=io.BytesIO())

            filename = os.path.basename(path)
            dirname = os.path.dirname(path)

            file_like = io.BytesIO(string.encode())

            info = tarfile.TarInfo(name=filename)
            info.size = len(file_like.getbuffer())

            tar.addfile(tarinfo=info, fileobj=file_like)

            self.container.put_archive(dirname, tar.fileobj.getvalue())

        else:
            self._files[path] = string

    def _deploy_routes(self):

        if any([isinstance(n, DockerSubnet) for n in self.subnets]) and not self.connect_to_internet:
            status_code, output = self.container.exec_run('ip route delete default')

            if status_code:
                print('Error({2}) while removing default docker route for {0}:\n{1}'.format(
                    self.name, output.decode(), status_code))

        for ip_range, gateway in self._routes.items():
            status_code, output = self.container.exec_run('ip route add {} via '.format(ip_range) + gateway)
            if status_code:
                print('Error({2}) while setting routes for {0}:\n{1}'.format(self.name, output.decode(), status_code))

    @classmethod
    def make_router(cls, *args, **kwargs):

        kwargs = kwargs.copy()
        kwargs['image'] = 'frrouting/frr'
        kwargs['entrypoint'] = '/usr/lib/frr/docker-start'
        # TODO: change to cap_add
        kwargs['privileged'] = True

        router = cls(*args, **kwargs)

        router.is_router = True

        return router

    def activate_protocol(self, protocol: str, config: str):
        daemon_name = protocol.lower() + 'd'
        cmd = "sed -i 's/{daemon}=no/{daemon}=yes/' /etc/frr/daemons &&".format(daemon=daemon_name)
        self._docker_kwargs['entrypoint'] = cmd + self._docker_kwargs['entrypoint']

        self.put_string('/etc/frr/{}.conf'.format(daemon_name), config)



from .base import BaseOverlay

config_template = '''router rip

{}

redistribute kernel
redistribute static
redistribute connected
'''


class RipOverlay(BaseOverlay):

    protocol = 'RIP'

    def configure_node(self, node):
        config = config_template.format('\n'.join(['neighbor {}'.format(s) for s in self._neighbors_ips(node)]))

        node.activate_protocol(self.protocol, config)



from ipaddress import ip_network, IPv4Network
from math import log2, ceil
from .. import defaults

import concurrent.futures
import traceback

import time


class Network:

    def __init__(self, ip_range: str or IPv4Network, local=False):
        self.ip_range = str(ip_range)
        self._available_subnets = [ip_network(ip_range) if type(ip_range) is str else ip_range, ]
        self._subnets = set()
        self._nodes = []
        self.is_local = local

        self.loaded = False
        self.started = False

    def add_node(self, node):
        self._nodes.append(node)
        node.add_network(self)

    def add_subnet(self, subnet):
        self._subnets.add(subnet)

    def create_subnet(self, subnet_name, *args, subnet_type: type = None, **kwargs):
        if subnet_type is None:
            subnet_type = defaults.subnet_class

        return subnet_type(subnet_name, *args, network=self, **kwargs)

    def get_subnet_range(self, size: int):
        max_prefix = 32 - ceil(log2(size + 2))
        #  size = 2^(32-p) - 2

        try:
            subnet = [net for net in self._available_subnets if net.prefixlen <= max_prefix][0]
        except IndexError:
            raise Exception('Network address pool is empty')

        self._available_subnets.remove(subnet)

        if subnet.prefixlen == max_prefix:
            return subnet

        try:
            smaller_subnet = next(subnet.subnets(new_prefix=max_prefix))
        except ValueError or StopIteration as e:
            raise Exception('Error while splitting net: {}'.format(e))

        self._available_subnets.extend(subnet.address_exclude(smaller_subnet))
        self._available_subnets.sort(key=lambda n: 32 - n.prefixlen)

        return smaller_subnet

    def start(self, interactive=False):
        if self.loaded:
            raise Exception('Network already loaded!')

        if not interactive:
            for subnet in self._subnets:
                subnet.start()

            for node in self._nodes:
                node.start()

            for node in self._nodes:
                # node.unlock()
                pass

            self.started = True
            self.loaded = True

        else:
            try:
                self.start(interactive=False)
            except:
                print('\nException while starting:\n_________________\n\n',
                      traceback.format_exc(), '\n_________________\n', sep='')

            i = ''
            while i not in ('y', 'n', 'yes', 'no'):
                i = input('Stop? (y/n): ')

            if i[0] == 'y':
                self.stop()

    def load(self, *args):
        for node in self._nodes:
            node.load(*args)

        for subnet in self._subnets:
            subnet.load(*args)

    def stop(self, *args):
        if not self.loaded:
            self.load(*args)

        for node in self._nodes:
            node.stop(*args)

        for subnet in self._subnets:
            subnet.stop(*args)

    def __contains__(self, item):
        return item in self._nodes or item in self._subnets


class NetworkConcurrent(Network):

    def __init__(self, *args, max_workers=None, **kwargs):
        self.max_workers = max_workers
        super().__init__(*args, **kwargs)

    def start(self, interactive=False):
        # TODO: debug & interactive realisation

        if self.loaded:
            raise Exception('Network already loaded!')

        if not interactive:
            for subnet in self._subnets:
                subnet.start()

            with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                for node in self._nodes:
                    executor.submit(node.start, dc=None)

            print('start time:', time.strftime('[%H:%M:%S]', time.gmtime()))

            # TODO: executor here too
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                for node in self._nodes:
                    executor.submit(node.unlock)
            '''
            for node in self._nodes:
                node.unlock()
            '''


            self.started = True
            self.loaded = True

        else:
            try:
                self.start(interactive=False)
            except:
                print('\nException while starting:\n_________________\n\n',
                      traceback.format_exc(), '\n_________________\n', sep='')

            i = ''
            while i not in ('y', 'n', 'yes', 'no'):
                i = input('Stop? (y/n): ')

            if i[0] == 'y':
                self.stop()

    def stop(self):
        if not self.loaded:
            self.load()

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            for node in self._nodes:
                executor.submit(node.stop, dc=None)

        for subnet in self._subnets:
            subnet.stop()



from .. import defaults
from .. import utils
from . import network as base_network
from . import node as base_node


class BaseSubnet:

    def __init__(self, name: str, *args, size: int = 1024, network: 'base_network.Network' = None):
        self.name = name
        self._network = network if network is not None else defaults.network
        self._network.add_subnet(self)
        self._max_size = size
        self._ip_range = self._network.get_subnet_range(self._max_size)
        self._hosts = list(self._ip_range.hosts())

        self._nodes_dict = {}
        self.started = False
        self.loaded = False

        for node in utils.args.list_from_args(args):
            self.add_node(node)

    def __iter__(self):
        return iter(self._nodes_dict)

    def _get_host_ip(self):
        try:
            return str(self._hosts.pop(0))
        except IndexError:
            raise Exception('Network {} is too small'.format(self.name))

    def add_node(self, node: 'base_node.BaseNode'):

        ip_address = self._get_host_ip()

        self._nodes_dict[ip_address] = node
        if node not in self._network:
            self._network.add_node(node)
        node.add_interface(ip_address, self)

    def start(self):
        raise NotImplementedError()

    def load(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def get_network(self):
        return self._network

    def get_gateway(self, node):
        router = None

        for ip, n in self._nodes_dict.items():

            if n is node:
                continue

            if n.is_gateway:
                return ip

            if not router and n.is_router:
                router = ip

        return router

    def get_nodes_dict(self):
        return self._nodes_dict.copy()

    def prefixlen(self):
        return self._ip_range.prefixlen

    def get_ip_range(self):
        return self._ip_range

    @property
    def nodes(self):
        return list(self._nodes_dict.values())

    def get_node_ip(self, node):
        for ip, n in self._nodes_dict.items():
            if n is node:
                return ip

        raise Exception('Node is not in subnet')

    def __repr__(self):
        return '{class_name}({name}, {ip_range})'.format(
            class_name=self.__class__.__name__,
            name=self.name,
            ip_range=self._ip_range
        )

    def __contains__(self, item):
        return item in self._nodes_dict.values()

    def __str__(self):
        return str(self._ip_range)



from .. import utils


class BaseOverlay:
    def __init__(self, *args):

        self._nodes = utils.args.list_from_args(args)

        for node in self._nodes:
            self.configure_node(node)

    def __contains__(self, item):
        return item in self._nodes

    def __iter__(self):
        return iter(self._nodes)

    def configure_node(self, node):
        raise NotImplementedError()

    def _node_subnets(self, node):
        """get node subnets that have other nodes from overlay"""
        return [s for s in node.subnets if any([n in s for n in self._nodes if n is not node])]

    def _neighbors_ips(self, node):

        ips = []

        for subnet in node.subnets:
            for ip, n in subnet.get_nodes_dict().items():
                if n in self and n is not node:
                    ips.append(ip)

        return ips



import os
import hashids

from .. import ldc
from .utils import my_hash, generate_wg_keys, ClusterHost, vpn_server_address
from .subnet import WgSubnet


node_config_template = '''[Interface]
PrivateKey = {private_key}

[Peer]
# vpn_server
Endpoint={vpn_server_address}:{vpn_server_port}
PublicKey = {vpn_server_public_key}
AllowedIPs = {network_ip_range}
PersistentKeepalive = 25
'''

node_startup_script = '''
pid=$(docker inspect -f '{{{{.State.Pid}}}}' {c_id}) && \\
mkdir -p /var/run/netns/ && \\
ln -sfT /proc/$pid/ns/net /var/run/netns/{c_id} && \\
ip link add dev {ifc} type wireguard && \\
wg setconf {ifc} {config_path} && \\
ip link set {ifc} netns {c_id} && \\
ip -n {c_id} addr add {ip_address}/{net_mask} dev {ifc} && \\
ip -n {c_id} link set {ifc} up
'''


class RemoteNode(ldc.DockerNode):

    def __init__(self, name, cluster_host: 'ClusterHost', **kwargs):
        self.cluster_host = cluster_host
        self.config_path = os.path.join('/tmp/l3ns', name) + '.conf'

        self.private_key, self.public_key = generate_wg_keys()

        super().__init__(name, **kwargs)

        self._client = cluster_host.get_docker_client()

    def _connect_subnet(self, network, ip):

        if isinstance(network, ldc.DockerSubnet):
            super()._connect_subnet(network, ip)
        elif isinstance(network, WgSubnet):
            # connect to wg_subnet

            self.cluster_host.upload_config(self.config_path, node_config_template.format(
                private_key=self.private_key,
                vpn_server_address=vpn_server_address,
                vpn_server_port=network.listen_port,
                vpn_server_public_key=network.public_key,
                network_ip_range=network.get_network().ip_range
            ))

            cmd = node_startup_script.format(
                ifc='wg-' + my_hash(self.name, network),
                ip_address=self.get_ip(network),
                net_mask=network.prefixlen(),
                config_path=self.config_path,
                c_id=self.container.short_id
            )

            # print('_________________\n', cmd, '_________________')

            status_code, stdout, stderr = self.cluster_host.exec_command(cmd)

            if status_code:
                print('Error while connecting node {} to the subnet {}:'.format(self.name, network.name))
                print('status_code: ' + str(status_code))
                if stdout:
                    print('stdout: ' + stdout)
                if stderr:
                    print('stderr: ' + stderr)
            else:
                print('{} connected to the subnet {}'.format(self.name, network.name))

        else:

            raise Exception("Unexpected network type: " + network.name)

    def __repr__(self):
        return '<{class_name}({name}@{hostname}, {ip})'.format(
            class_name=self.__class__.__name__,
            name=self.name,
            ip=self.get_ip(),
            hostname=self.cluster_host.address)



import setuptools


with open("PYPI_README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='l3ns',
    version='0.0.1',
    license='MIT',
    description='L3 Network Simulator',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Daniil Malevanniy',
    author_email='rukmarr@gmail.com',
    url='https://github.com/rukmarr/l3ns',
    packages=setuptools.find_packages(),
    install_requires=[
          'docker',
          'hashids',
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
)



import subprocess
import socket
import time
import os
from contextlib import closing

import docker
import paramiko

# TODO: move to config
vpn_server_address = 't1'

known_hosts_path = os.path.expanduser(os.path.join("~", ".ssh", "known_hosts"))
private_key_path = os.path.expanduser(os.path.join("~", ".ssh", "id_rsa"))


def generate_wg_keys():
    cmd = 'wg genkey | tee /dev/stderr | wg pubkey'

    ret = subprocess.run(cmd, shell=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         universal_newlines=True)

    return ret.stderr.strip(), ret.stdout.strip()


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


class ClusterHost:

    def __init__(self, address, login='root', port=22):
        self.address = address
        self.port = port
        self.login = login
        self.docker_port = find_free_port()

        cmd = 'ssh -p {port} -NL {docker_port}:/var/run/docker.sock {login}@{address}'.format(
            docker_port=self.docker_port,
            login=login,
            address=address,
            port=port
        )

        self.forwarder = subprocess.Popen(cmd,
                                          shell=True,
                                          stdout=subprocess.PIPE,
                                          stderr=subprocess.STDOUT,
                                          stdin=subprocess.PIPE,
                                          universal_newlines=True)

        self.ssh = paramiko.SSHClient()
        self.ssh.load_host_keys(known_hosts_path)
        self.ssh.connect(self.address, port=self.port, username=self.login, key_filename=private_key_path)

        self.sftp = self.ssh.open_sftp()

        timeout = 60
        while not self.test_client() and timeout > 0:
            time.sleep(0.5)
            timeout -= 0.5

        self._client = self.get_docker_client(cached=False)


    def get_docker_client(self, cached=True) -> docker.DockerClient:
        if cached:
            return self._client
        else:
            return docker.DockerClient(base_url='tcp://127.0.0.1:{}'.format(self.docker_port))

    def test_client(self):

        try:
            self.get_docker_client(cached=False).version()
            return True
        except:
            return False

    def upload_config(self, path, text):
        with self.sftp.open(os.path.join('/tmp/madt', path), 'w') as config_file:
            config_file.write(text)

    def exec_command(self, cmd):

        stdin, stdout, stderr = self.ssh.exec_command(cmd)

        return stdout.channel.recv_exit_status(), stdout.read().decode(), stderr.read().decode()


def my_hash(*args):
    return str(hash(tuple(args)) % 10 ** 8)
